"""
Adagrad and its variants implementation.
"""

import torch
import torch.optim as optim
import math
from typing import Optional, Callable, Union, List, Dict, Any
from ...utils.registry import register_optimizer


@register_optimizer("adagrad")
class Adagrad(optim.Adagrad):
    """Adagrad optimizer with enhanced features."""
    
    def __init__(
        self,
        params: Union[List[torch.Tensor], Dict[str, Any]],
        lr: float = 1e-2,
        lr_decay: float = 0,
        weight_decay: float = 0,
        initial_accumulator_value: float = 0,
        eps: float = 1e-10,
        **kwargs
    ):
        super().__init__(params, lr, lr_decay, weight_decay, initial_accumulator_value, eps)


@register_optimizer("adadelta")
class Adadelta(optim.Adadelta):
    """Adadelta optimizer with enhanced features."""
    
    def __init__(
        self,
        params: Union[List[torch.Tensor], Dict[str, Any]],
        lr: float = 1.0,
        rho: float = 0.9,
        eps: float = 1e-6,
        weight_decay: float = 0,
        **kwargs
    ):
        super().__init__(params, lr, rho, eps, weight_decay)


@register_optimizer("adafactor")
class AdaFactor(optim.Optimizer):
    """
    AdaFactor: Adaptive Learning Rates with Sublinear Memory Cost.
    
    Reference: https://arxiv.org/abs/1804.04235
    """
    
    def __init__(
        self,
        params: Union[List[torch.Tensor], Dict[str, Any]],
        lr: Optional[float] = None,
        eps2: float = 1e-30,
        cliping_threshold: float = 1.0,
        decay_rate: float = -0.8,
        beta1: Optional[float] = None,
        weight_decay: float = 0.0,
        scale_parameter: bool = True,
        relative_step: bool = True,
        **kwargs
    ):
        if lr is not None and relative_step:
            raise ValueError("Cannot combine manual lr and relative_step options")
        if not relative_step and lr is None:
            raise ValueError("Must set lr if relative_step is False")
        
        defaults = dict(
            lr=lr, eps2=eps2, cliping_threshold=cliping_threshold, decay_rate=decay_rate,
            beta1=beta1, weight_decay=weight_decay, scale_parameter=scale_parameter,
            relative_step=relative_step
        )
        super().__init__(params, defaults)
    
    def _get_lr(self, param_group, param_state):
        """Compute learning rate for each parameter tensor."""
        min_step = 1e-6 * param_state['step'] if param_group['scale_parameter'] else 1e-2
        rel_step_sz = min(min_step, 1.0 / math.sqrt(param_state['step']))
        param_scale = 1.0
        if param_group['scale_parameter']:
            param_scale = max(param_group['eps2'], param_state['RMS'])
        return param_scale * rel_step_sz
    
    def _get_options(self, param_group, param_shape):
        """Determine factorization options."""
        factored = len(param_shape) >= 2
        use_first_moment = param_group['beta1'] is not None
        return factored, use_first_moment
    
    def _rms(self, tensor):
        """Root mean square."""
        return tensor.norm(2) / (tensor.numel() ** 0.5)
    
    def _approx_sq_grad(self, exp_avg_sq_row, exp_avg_sq_col):
        """Approximation of exponential moving average of square of gradient."""
        r_factor = ((exp_avg_sq_row / exp_avg_sq_row.mean(dim=-1, keepdim=True))
                   .rsqrt_().unsqueeze(-1).clamp_(0, math.inf))
        c_factor = (exp_avg_sq_col.rsqrt()).unsqueeze(0).clamp_(0, math.inf)
        return torch.mul(r_factor, c_factor)
    
    def step(self, closure: Optional[Callable] = None):
        """Performs a single optimization step."""
        loss = None
        if closure is not None:
            loss = closure()
        
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                
                grad = p.grad.data
                if grad.dtype in {torch.float16, torch.bfloat16}:
                    grad = grad.float()
                
                state = self.state[p]
                grad_shape = grad.shape
                
                factored, use_first_moment = self._get_options(group, grad_shape)
                
                # State Initialization
                if len(state) == 0:
                    state['step'] = 0
                    
                    if use_first_moment:
                        # Exponential moving average of gradient values
                        state['exp_avg'] = torch.zeros_like(grad).float()
                    if factored:
                        state['exp_avg_sq_row'] = torch.zeros(grad_shape[:-1]).float()
                        state['exp_avg_sq_col'] = torch.zeros(grad_shape[:-2] + grad_shape[-1:]).float()
                    else:
                        state['exp_avg_sq'] = torch.zeros_like(grad).float()
                    
                    state['RMS'] = 0
                
                p_data_fp32 = p.data.float() if p.data.dtype in {torch.float16, torch.bfloat16} else p.data
                
                state['step'] += 1
                state['RMS'] = self._rms(p_data_fp32)
                
                lr = group['lr']
                if group['lr'] is None:
                    lr = self._get_lr(group, state)
                
                beta2t = 1.0 - math.pow(state['step'], group['decay_rate'])
                update = grad**2 + group['eps2']
                
                if factored:
                    exp_avg_sq_row = state['exp_avg_sq_row']
                    exp_avg_sq_col = state['exp_avg_sq_col']
                    
                    exp_avg_sq_row.mul_(beta2t).add_(update.mean(dim=-1), alpha=1.0-beta2t)
                    exp_avg_sq_col.mul_(beta2t).add_(update.mean(dim=-2), alpha=1.0-beta2t)
                    update = self._approx_sq_grad(exp_avg_sq_row, exp_avg_sq_col)
                    update.mul_(grad)
                else:
                    exp_avg_sq = state['exp_avg_sq']
                    exp_avg_sq.mul_(beta2t).add_(update, alpha=1.0-beta2t)
                    update = exp_avg_sq.rsqrt().mul_(grad)
                
                update.div_((self._rms(update) / group['cliping_threshold']).clamp_(min=1.0))
                
                if use_first_moment:
                    exp_avg = state['exp_avg']
                    exp_avg.mul_(group['beta1']).add_(update, alpha=1 - group['beta1'])
                    update = exp_avg
                
                if group['weight_decay'] != 0:
                    p_data_fp32.mul_(1 - group['weight_decay'] * lr)
                
                p_data_fp32.add_(update, alpha=-lr)
                
                if p.data.dtype in {torch.float16, torch.bfloat16}:
                    p.data.copy_(p_data_fp32)
        
        return loss


@register_optimizer("adagc")
class AdaGC(optim.Optimizer):
    """AdaGC: Adaptive Gradient Compression."""
    
    def __init__(
        self,
        params: Union[List[torch.Tensor], Dict[str, Any]],
        lr: float = 1e-3,
        betas: tuple = (0.9, 0.999),
        eps: float = 1e-8,
        weight_decay: float = 0,
        **kwargs
    ):
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)
        super().__init__(params, defaults)
    
    def step(self, closure: Optional[Callable] = None):
        """Performs a single optimization step."""
        loss = None
        if closure is not None:
            loss = closure()
        
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                
                grad = p.grad.data
                state = self.state[p]
                
                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p.data)
                    state['exp_avg_sq'] = torch.zeros_like(p.data)
                
                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']
                
                state['step'] += 1
                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']
                
                # Gradient compression
                grad_norm = torch.norm(grad)
                if grad_norm > 1.0:
                    grad = grad / grad_norm
                
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                
                denom = (exp_avg_sq / bias_correction2).sqrt().add_(group['eps'])
                
                if group['weight_decay'] != 0:
                    p.data.mul_(1 - group['lr'] * group['weight_decay'])
                
                p.data.addcdiv_(exp_avg, denom, value=-group['lr'] / bias_correction1)
        
        return loss


@register_optimizer("adago")
class AdaGO(optim.Optimizer):
    """AdaGO: Adaptive Gradient with Orthogonalization."""
    
    def __init__(
        self,
        params: Union[List[torch.Tensor], Dict[str, Any]],
        lr: float = 1e-3,
        betas: tuple = (0.9, 0.999),
        eps: float = 1e-8,
        weight_decay: float = 0,
        **kwargs
    ):
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)
        super().__init__(params, defaults)
    
    def step(self, closure: Optional[Callable] = None):
        """Performs a single optimization step."""
        loss = None
        if closure is not None:
            loss = closure()
        
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                
                grad = p.grad.data
                state = self.state[p]
                
                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p.data)
                    state['exp_avg_sq'] = torch.zeros_like(p.data)
                    state['prev_grad'] = torch.zeros_like(p.data)
                
                exp_avg, exp_avg_sq, prev_grad = state['exp_avg'], state['exp_avg_sq'], state['prev_grad']
                beta1, beta2 = group['betas']
                
                state['step'] += 1
                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']
                
                # Orthogonalize gradient
                dot_product = torch.sum(grad * prev_grad)
                prev_norm_sq = torch.sum(prev_grad * prev_grad)
                if prev_norm_sq > 0:
                    grad = grad - (dot_product / prev_norm_sq) * prev_grad
                
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                
                denom = (exp_avg_sq / bias_correction2).sqrt().add_(group['eps'])
                
                if group['weight_decay'] != 0:
                    p.data.mul_(1 - group['lr'] * group['weight_decay'])
                
                p.data.addcdiv_(exp_avg, denom, value=-group['lr'] / bias_correction1)
                
                # Update previous gradient
                prev_grad.copy_(grad)
        
        return loss


@register_optimizer("adalomo")
class AdaLOMO(optim.Optimizer):
    """AdaLOMO: Low-Memory Optimization for Large Models."""
    
    def __init__(
        self,
        model: torch.nn.Module,
        lr: float = 1e-3,
        eps: float = 1e-8,
        clip_threshold: float = 1.0,
        **kwargs
    ):
        self.model = model
        params = list(model.parameters())
        defaults = dict(lr=lr, eps=eps, clip_threshold=clip_threshold)
        super().__init__(params, defaults)
    
    def step(self, closure: Optional[Callable] = None):
        """Performs a single optimization step."""
        loss = None
        if closure is not None:
            loss = closure()
        
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                
                grad = p.grad.data
                state = self.state[p]
                
                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg_sq'] = torch.zeros_like(p.data)
                
                exp_avg_sq = state['exp_avg_sq']
                state['step'] += 1
                
                # Update second moment
                exp_avg_sq.mul_(0.999).addcmul_(grad, grad, value=0.001)
                
                # Compute update
                bias_correction = 1 - 0.999 ** state['step']
                denom = (exp_avg_sq / bias_correction).sqrt().add_(group['eps'])
                
                # Clip gradient
                grad_norm = torch.norm(grad)
                if grad_norm > group['clip_threshold']:
                    grad = grad * (group['clip_threshold'] / grad_norm)
                
                p.data.addcdiv_(grad, denom, value=-group['lr'])
        
        return loss


@register_optimizer("adai")
class Adai(optim.Optimizer):
    """Adai: Disentangling the Effects of Adaptive Learning Rate and Momentum."""
    
    def __init__(
        self,
        params: Union[List[torch.Tensor], Dict[str, Any]],
        lr: float = 1e-3,
        betas: tuple = (0.9, 0.999),
        eps: float = 1e-8,
        weight_decay: float = 0,
        **kwargs
    ):
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)
        super().__init__(params, defaults)
    
    def step(self, closure: Optional[Callable] = None):
        """Performs a single optimization step."""
        loss = None
        if closure is not None:
            loss = closure()
        
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                
                grad = p.grad.data
                state = self.state[p]
                
                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p.data)
                    state['exp_avg_sq'] = torch.zeros_like(p.data)
                
                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']
                
                state['step'] += 1
                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']
                
                # Adaptive inertia
                inertia = beta1 * (1 - beta1 ** (state['step'] - 1)) / bias_correction1
                
                exp_avg.mul_(inertia).add_(grad, alpha=1 - inertia)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                
                denom = (exp_avg_sq / bias_correction2).sqrt().add_(group['eps'])
                
                if group['weight_decay'] != 0:
                    p.data.mul_(1 - group['lr'] * group['weight_decay'])
                
                p.data.addcdiv_(exp_avg, denom, value=-group['lr'])
        
        return loss


# Simplified implementations for remaining variants
@register_optimizer("adalite")
class Adalite(optim.Optimizer):
    """Adalite: Lightweight Adaptive Optimizer."""
    
    def __init__(self, params, lr=1e-3, eps=1e-8, weight_decay=0, **kwargs):
        defaults = dict(lr=lr, eps=eps, weight_decay=weight_decay)
        super().__init__(params, defaults)
    
    def step(self, closure=None):
        loss = None
        if closure is not None:
            loss = closure()
        
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                
                grad = p.grad.data
                state = self.state[p]
                
                if len(state) == 0:
                    state['sum_sq'] = torch.zeros_like(p.data)
                
                sum_sq = state['sum_sq']
                sum_sq.addcmul_(grad, grad)
                
                std = sum_sq.sqrt().add_(group['eps'])
                
                if group['weight_decay'] != 0:
                    p.data.mul_(1 - group['lr'] * group['weight_decay'])
                
                p.data.addcdiv_(grad, std, value=-group['lr'])
        
        return loss


@register_optimizer("adammini")
class AdamMini(optim.Optimizer):
    """AdamMini: Minimal memory footprint Adam."""
    
    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0, **kwargs):
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)
        super().__init__(params, defaults)
    
    def step(self, closure=None):
        loss = None
        if closure is not None:
            loss = closure()
        
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                
                grad = p.grad.data
                state = self.state[p]
                
                if len(state) == 0:
                    state['step'] = 0
                    state['v'] = torch.zeros_like(p.data)
                
                v = state['v']
                beta1, beta2 = group['betas']
                
                state['step'] += 1
                bias_correction = 1 - beta2 ** state['step']
                
                v.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                denom = (v / bias_correction).sqrt().add_(group['eps'])
                
                if group['weight_decay'] != 0:
                    p.data.mul_(1 - group['lr'] * group['weight_decay'])
                
                p.data.addcdiv_(grad, denom, value=-group['lr'])
        
        return loss


@register_optimizer("adamod")
class AdaMod(optim.Optimizer):
    """AdaMod: An Adaptive and Momental Bound Method."""
    
    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0, **kwargs):
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)
        super().__init__(params, defaults)
    
    def step(self, closure=None):
        loss = None
        if closure is not None:
            loss = closure()
        
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                
                grad = p.grad.data
                state = self.state[p]
                
                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p.data)
                    state['exp_avg_sq'] = torch.zeros_like(p.data)
                
                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']
                
                state['step'] += 1
                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']
                
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                
                # Momental bound
                denom = (exp_avg_sq / bias_correction2).sqrt().add_(group['eps'])
                step_size = group['lr'] / bias_correction1
                
                # Adaptive bound
                bound = step_size * torch.ones_like(denom)
                bound = torch.min(bound, 1.0 / denom)
                
                if group['weight_decay'] != 0:
                    p.data.mul_(1 - group['lr'] * group['weight_decay'])
                
                p.data.add_(exp_avg, alpha=-bound)
        
        return loss


@register_optimizer("adanorm")
class AdaNorm(optim.Optimizer):
    """AdaNorm: Adaptive Gradient Norm Correction."""
    
    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0, **kwargs):
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)
        super().__init__(params, defaults)
    
    def step(self, closure=None):
        loss = None
        if closure is not None:
            loss = closure()
        
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                
                grad = p.grad.data
                state = self.state[p]
                
                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p.data)
                    state['exp_avg_sq'] = torch.zeros_like(p.data)
                
                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']
                
                state['step'] += 1
                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']
                
                # Normalize gradient
                grad_norm = torch.norm(grad)
                if grad_norm > 0:
                    grad = grad / grad_norm
                
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                
                denom = (exp_avg_sq / bias_correction2).sqrt().add_(group['eps'])
                
                if group['weight_decay'] != 0:
                    p.data.mul_(1 - group['lr'] * group['weight_decay'])
                
                p.data.addcdiv_(exp_avg, denom, value=-group['lr'] / bias_correction1)
        
        return loss


@register_optimizer("adapnm")
class AdaPNM(optim.Optimizer):
    """AdaPNM: Adaptive Positive-Negative Momentum."""
    
    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0, **kwargs):
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)
        super().__init__(params, defaults)
    
    def step(self, closure=None):
        loss = None
        if closure is not None:
            loss = closure()
        
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                
                grad = p.grad.data
                state = self.state[p]
                
                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg_pos'] = torch.zeros_like(p.data)
                    state['exp_avg_neg'] = torch.zeros_like(p.data)
                    state['exp_avg_sq'] = torch.zeros_like(p.data)
                
                exp_avg_pos, exp_avg_neg, exp_avg_sq = (
                    state['exp_avg_pos'], state['exp_avg_neg'], state['exp_avg_sq']
                )
                beta1, beta2 = group['betas']
                
                state['step'] += 1
                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']
                
                # Separate positive and negative gradients
                grad_pos = torch.clamp(grad, min=0)
                grad_neg = torch.clamp(grad, max=0)
                
                exp_avg_pos.mul_(beta1).add_(grad_pos, alpha=1 - beta1)
                exp_avg_neg.mul_(beta1).add_(grad_neg, alpha=1 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                
                # Combine positive and negative momentum
                exp_avg = exp_avg_pos + exp_avg_neg
                
                denom = (exp_avg_sq / bias_correction2).sqrt().add_(group['eps'])
                
                if group['weight_decay'] != 0:
                    p.data.mul_(1 - group['lr'] * group['weight_decay'])
                
                p.data.addcdiv_(exp_avg, denom, value=-group['lr'] / bias_correction1)
        
        return loss



"""
SGD and its variants implementation.
"""

import torch
import torch.optim as optim
import math
from typing import Optional, Callable, Union, List, Dict, Any
from ...utils.registry import register_optimizer


@register_optimizer("sgd")
class SGD(optim.SGD):
    """SGD optimizer with enhanced features."""
    
    def __init__(
        self,
        params: Union[List[torch.Tensor], Dict[str, Any]],
        lr: float = 1e-3,
        momentum: float = 0,
        dampening: float = 0,
        weight_decay: float = 0,
        nesterov: bool = False,
        **kwargs
    ):
        super().__init__(params, lr, momentum, dampening, weight_decay, nesterov)


@register_optimizer("nesterovsgd")
class NesterovSGD(optim.SGD):
    """Nesterov SGD optimizer."""
    
    def __init__(
        self,
        params: Union[List[torch.Tensor], Dict[str, Any]],
        lr: float = 1e-3,
        momentum: float = 0.9,
        dampening: float = 0,
        weight_decay: float = 0,
        **kwargs
    ):
        super().__init__(params, lr, momentum, dampening, weight_decay, nesterov=True)


@register_optimizer("qhm")
class QHM(optim.Optimizer):
    """
    QHM: Quasi-Hyperbolic Momentum optimizer.
    
    Reference: https://arxiv.org/abs/1810.06801
    """
    
    def __init__(
        self,
        params: Union[List[torch.Tensor], Dict[str, Any]],
        lr: float = 1e-3,
        momentum: float = 0.999,
        nu: float = 0.7,
        weight_decay: float = 0,
        weight_decouple: bool = True,
        **kwargs
    ):
        defaults = dict(
            lr=lr, momentum=momentum, nu=nu, weight_decay=weight_decay,
            weight_decouple=weight_decouple
        )
        super().__init__(params, defaults)
    
    def step(self, closure: Optional[Callable] = None):
        """Performs a single optimization step."""
        loss = None
        if closure is not None:
            loss = closure()
        
        for group in self.param_groups:
            weight_decay = group['weight_decay']
            momentum = group['momentum']
            
            for p in group['params']:
                if p.grad is None:
                    continue
                
                param_state = self.state[p]
                d_p = p.grad.data
                
                if len(param_state) == 0:
                    param_state['momentum_buffer'] = torch.zeros_like(p.data)
                
                buf = param_state['momentum_buffer']
                
                if weight_decay != 0:
                    if group['weight_decouple']:
                        # Decoupled weight decay (L2 regularization)
                        p.data.mul_(1.0 - group['lr'] * weight_decay)
                    else:
                        # L2 weight decay
                        d_p = d_p.add(p.data, alpha=weight_decay)
                
                # Update biased first moment estimate
                buf.mul_(momentum).add_(d_p)
                
                # Update parameters
                p.data.add_(d_p, alpha=-group['lr'] * group['nu'])
                p.data.add_(buf, alpha=-group['lr'] * (1.0 - group['nu']))
        
        return loss


@register_optimizer("aggmo")
class AggMo(optim.Optimizer):
    """
    AggMo: Aggregated Momentum optimizer.
    
    Reference: https://arxiv.org/abs/1804.00325
    """
    
    def __init__(
        self,
        params: Union[List[torch.Tensor], Dict[str, Any]],
        lr: float = 1e-3,
        betas: List[float] = [0.0, 0.9, 0.99],
        weight_decay: float = 0,
        **kwargs
    ):
        defaults = dict(lr=lr, betas=betas, weight_decay=weight_decay)
        super().__init__(params, defaults)
    
    def step(self, closure: Optional[Callable] = None):
        """Performs a single optimization step."""
        loss = None
        if closure is not None:
            loss = closure()
        
        for group in self.param_groups:
            weight_decay = group['weight_decay']
            betas = group['betas']
            
            for p in group['params']:
                if p.grad is None:
                    continue
                
                param_state = self.state[p]
                d_p = p.grad.data
                
                if weight_decay != 0:
                    d_p = d_p.add(p.data, alpha=weight_decay)
                
                if len(param_state) == 0:
                    param_state['momentum_buffers'] = [
                        torch.zeros_like(p.data) for _ in betas
                    ]
                
                buffers = param_state['momentum_buffers']
                
                # Update momentum buffers
                for i, beta in enumerate(betas):
                    buffers[i].mul_(beta).add_(d_p)
                
                # Aggregate momentum
                avg_momentum = sum(buffers) / len(buffers)
                
                # Update parameters
                p.data.add_(avg_momentum, alpha=-group['lr'])
        
        return loss


@register_optimizer("swats")
class SWATS(optim.Optimizer):
    """
    SWATS: Switching from Adam to SGD optimizer.
    
    Reference: https://arxiv.org/abs/1712.07628
    """
    
    def __init__(
        self,
        params: Union[List[torch.Tensor], Dict[str, Any]],
        lr: float = 1e-3,
        betas: tuple = (0.9, 0.999),
        eps: float = 1e-8,
        weight_decay: float = 0,
        amsgrad: bool = False,
        **kwargs
    ):
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay, amsgrad=amsgrad)
        super().__init__(params, defaults)
    
    def step(self, closure: Optional[Callable] = None):
        """Performs a single optimization step."""
        loss = None
        if closure is not None:
            loss = closure()
        
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError('SWATS does not support sparse gradients')
                
                state = self.state[p]
                
                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p.data)
                    state['exp_avg_sq'] = torch.zeros_like(p.data)
                    if group['amsgrad']:
                        state['max_exp_avg_sq'] = torch.zeros_like(p.data)
                    state['switched'] = False
                
                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                if group['amsgrad']:
                    max_exp_avg_sq = state['max_exp_avg_sq']
                beta1, beta2 = group['betas']
                
                state['step'] += 1
                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']
                
                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                
                if group['amsgrad']:
                    # Maintains the maximum of all 2nd moment running avg. till now
                    torch.max(max_exp_avg_sq, exp_avg_sq, out=max_exp_avg_sq)
                    # Use the max. for normalizing running avg. of gradient
                    denom = (max_exp_avg_sq / bias_correction2).sqrt().add_(group['eps'])
                else:
                    denom = (exp_avg_sq / bias_correction2).sqrt().add_(group['eps'])
                
                # Check switching condition
                if state['step'] > 100 and not state['switched']:
                    # Compute projection
                    proj = torch.sum(exp_avg * grad) / (torch.norm(exp_avg) * torch.norm(grad) + 1e-8)
                    if proj.abs() > 0.9:  # High correlation, switch to SGD
                        state['switched'] = True
                
                # Apply weight decay
                if group['weight_decay'] != 0:
                    p.data.mul_(1 - group['lr'] * group['weight_decay'])
                
                # Update parameters
                if state['switched']:
                    # Use SGD update
                    p.data.add_(grad, alpha=-group['lr'])
                else:
                    # Use Adam update
                    p.data.addcdiv_(exp_avg, denom, value=-group['lr'] / bias_correction1)
        
        return loss


@register_optimizer("sgdp")
class SGDP(optim.Optimizer):
    """
    SGDP: SGD with Projection optimizer.
    
    Reference: https://arxiv.org/abs/2006.08217
    """
    
    def __init__(
        self,
        params: Union[List[torch.Tensor], Dict[str, Any]],
        lr: float = 1e-3,
        momentum: float = 0,
        eps: float = 1e-8,
        delta: float = 0.1,
        wd_ratio: float = 0.1,
        nesterov: bool = False,
        weight_decay: float = 0,
        **kwargs
    ):
        defaults = dict(
            lr=lr, momentum=momentum, eps=eps, delta=delta, wd_ratio=wd_ratio,
            nesterov=nesterov, weight_decay=weight_decay
        )
        super().__init__(params, defaults)
    
    def step(self, closure: Optional[Callable] = None):
        """Performs a single optimization step."""
        loss = None
        if closure is not None:
            loss = closure()
        
        for group in self.param_groups:
            weight_decay = group['weight_decay']
            momentum = group['momentum']
            
            for p in group['params']:
                if p.grad is None:
                    continue
                
                param_state = self.state[p]
                d_p = p.grad.data
                
                if weight_decay != 0:
                    d_p = d_p.add(p.data, alpha=weight_decay)
                
                if momentum != 0:
                    if 'momentum_buffer' not in param_state:
                        buf = param_state['momentum_buffer'] = torch.zeros_like(p.data)
                        buf.mul_(momentum).add_(d_p)
                    else:
                        buf = param_state['momentum_buffer']
                        buf.mul_(momentum).add_(d_p, alpha=1)
                    
                    if group['nesterov']:
                        d_p = d_p.add(buf, alpha=momentum)
                    else:
                        d_p = buf
                
                # Projection
                wd_ratio = 1.0
                if len(p.shape) > 1:
                    cosine_sim = torch.sum(d_p * p.data) / (torch.norm(d_p) * torch.norm(p.data) + group['eps'])
                    
                    if cosine_sim.abs() < group['delta']:
                        # Project gradient
                        proj_norm = torch.sum(d_p * p.data) / (torch.norm(p.data) ** 2 + group['eps'])
                        d_p = d_p - proj_norm * p.data
                        wd_ratio = group['wd_ratio']
                
                # Update parameters
                p.data.add_(d_p, alpha=-group['lr'])
                
                # Apply projected weight decay
                if weight_decay != 0:
                    p.data.mul_(1 - group['lr'] * group['weight_decay'] * wd_ratio)
        
        return loss


@register_optimizer("sgdsai")
class SGDSaI(optim.Optimizer):
    """SGDSaI: SGD with Scale-adaptive and Inertial momentum."""
    
    def __init__(
        self,
        params: Union[List[torch.Tensor], Dict[str, Any]],
        lr: float = 1e-3,
        momentum: float = 0.9,
        weight_decay: float = 0,
        **kwargs
    ):
        defaults = dict(lr=lr, momentum=momentum, weight_decay=weight_decay)
        super().__init__(params, defaults)
    
    def step(self, closure: Optional[Callable] = None):
        """Performs a single optimization step."""
        loss = None
        if closure is not None:
            loss = closure()
        
        for group in self.param_groups:
            weight_decay = group['weight_decay']
            momentum = group['momentum']
            
            for p in group['params']:
                if p.grad is None:
                    continue
                
                param_state = self.state[p]
                d_p = p.grad.data
                
                if weight_decay != 0:
                    d_p = d_p.add(p.data, alpha=weight_decay)
                
                if len(param_state) == 0:
                    param_state['momentum_buffer'] = torch.zeros_like(p.data)
                    param_state['step'] = 0
                
                buf = param_state['momentum_buffer']
                param_state['step'] += 1
                
                # Scale-adaptive inertial momentum
                scale = 1.0 / (1.0 + param_state['step'] * 0.001)
                adaptive_momentum = momentum * scale
                
                buf.mul_(adaptive_momentum).add_(d_p)
                
                # Update parameters
                p.data.add_(buf, alpha=-group['lr'])
        
        return loss


@register_optimizer("signsgd")
class SignSGD(optim.Optimizer):
    """
    SignSGD: Sign SGD optimizer.
    
    Reference: https://arxiv.org/abs/1802.04434
    """
    
    def __init__(
        self,
        params: Union[List[torch.Tensor], Dict[str, Any]],
        lr: float = 1e-3,
        momentum: float = 0,
        weight_decay: float = 0,
        **kwargs
    ):
        defaults = dict(lr=lr, momentum=momentum, weight_decay=weight_decay)
        super().__init__(params, defaults)
    
    def step(self, closure: Optional[Callable] = None):
        """Performs a single optimization step."""
        loss = None
        if closure is not None:
            loss = closure()
        
        for group in self.param_groups:
            weight_decay = group['weight_decay']
            momentum = group['momentum']
            
            for p in group['params']:
                if p.grad is None:
                    continue
                
                param_state = self.state[p]
                d_p = p.grad.data
                
                if weight_decay != 0:
                    d_p = d_p.add(p.data, alpha=weight_decay)
                
                # Sign of gradient
                d_p = torch.sign(d_p)
                
                if momentum != 0:
                    if 'momentum_buffer' not in param_state:
                        buf = param_state['momentum_buffer'] = torch.zeros_like(p.data)
                        buf.mul_(momentum).add_(d_p)
                    else:
                        buf = param_state['momentum_buffer']
                        buf.mul_(momentum).add_(d_p)
                    
                    d_p = buf
                
                # Update parameters
                p.data.add_(d_p, alpha=-group['lr'])
        
        return loss



import torch
import torch.nn as nn
from torch.optim.optimizer import Optimizer
from typing import Any, Dict, List, Optional, Tuple
import math
from collections import defaultdict, deque

class LBFGS(Optimizer):
    """Limited-memory Broyden-Fletcher-Goldfarb-Shanno optimizer"""
    def __init__(self, params, lr=1, max_iter=20, max_eval=None, tolerance_grad=1e-7, 
                 tolerance_change=1e-9, history_size=100, line_search_fn=None):
        if max_eval is None:
            max_eval = max_iter * 5 // 4
        defaults = dict(lr=lr, max_iter=max_iter, max_eval=max_eval, 
                       tolerance_grad=tolerance_grad, tolerance_change=tolerance_change,
                       history_size=history_size, line_search_fn=line_search_fn)
        super().__init__(params, defaults)

        if len(self.param_groups) != 1:
            raise ValueError("LBFGS doesn't support per-parameter options")

        self._params = self.param_groups[0]['params']
        self._numel_cache = None

    def _numel(self):
        if self._numel_cache is None:
            self._numel_cache = sum(p.numel() for p in self._params)
        return self._numel_cache

    def _gather_flat_grad(self):
        views = []
        for p in self._params:
            if p.grad is None:
                view = p.new(p.numel()).zero_()
            elif p.grad.is_sparse:
                view = p.grad.to_dense().view(-1)
            else:
                view = p.grad.view(-1)
            views.append(view)
        return torch.cat(views, 0)

    def _add_grad(self, step_size, update):
        offset = 0
        for p in self._params:
            numel = p.numel()
            p.add_(update[offset:offset+numel].view_as(p), alpha=step_size)
            offset += numel

    def step(self, closure):
        """Perform a single optimization step"""
        assert len(self.param_groups) == 1
        group = self.param_groups[0]
        lr = group['lr']
        max_iter = group['max_iter']
        max_eval = group['max_eval']
        tolerance_grad = group['tolerance_grad']
        tolerance_change = group['tolerance_change']
        line_search_fn = group['line_search_fn']
        history_size = group['history_size']

        state = self.state[self._params[0]]
        
        # initialize state
        if len(state) == 0:
            state['func_evals'] = 0
            state['n_iter'] = 0

        # evaluate initial f(x) and df/dx
        orig_loss = closure()
        loss = float(orig_loss)
        current_evals = 1
        state['func_evals'] += 1

        flat_grad = self._gather_flat_grad()
        opt_cond = flat_grad.abs().max() <= tolerance_grad

        # check optimality condition
        if opt_cond:
            return orig_loss

        # tensors cached in state (for tracing)
        d = state.get('d')
        t = state.get('t')
        old_dirs = state.get('old_dirs')
        old_stps = state.get('old_stps')
        ro = state.get('ro')
        H_diag = state.get('H_diag')
        prev_flat_grad = state.get('prev_flat_grad')
        prev_loss = state.get('prev_loss')

        n_iter = 0
        # optimize for a max of max_iter iterations
        while n_iter < max_iter:
            # keep track of nb of iterations
            n_iter += 1
            state['n_iter'] += 1

            ############################################################
            # compute gradient descent direction
            ############################################################
            if state['n_iter'] == 1:
                d = flat_grad.neg()
                old_dirs = []
                old_stps = []
                ro = []
                H_diag = 1
            else:
                # do lbfgs update (update memory)
                y = flat_grad.sub(prev_flat_grad)
                s = d.mul(t)
                ys = y.dot(s)  # y*s
                if ys > 1e-10:
                    # updating memory
                    if len(old_dirs) == history_size:
                        # shift history by one (limited-memory)
                        old_dirs.pop(0)
                        old_stps.pop(0)
                        ro.pop(0)

                    # store new direction/step
                    old_dirs.append(y)
                    old_stps.append(s)
                    ro.append(1. / ys)

                    # update scale of initial Hessian approximation
                    H_diag = ys / y.dot(y)  # (y*s)/(y*y)

                # compute the approximate (L-BFGS) inverse Hessian
                # multiplied by the gradient
                num_old = len(old_dirs)

                if 'al' not in state:
                    state['al'] = [None] * history_size
                al = state['al']

                # iteration in L-BFGS loop collapsed to use just one buffer
                q = flat_grad.neg()
                for i in range(num_old - 1, -1, -1):
                    al[i] = old_stps[i].dot(q) * ro[i]
                    q.add_(old_dirs[i], alpha=-al[i])

                # multiply by initial Hessian
                # r/d is the final direction
                d = r = torch.mul(q, H_diag)
                for i in range(num_old):
                    be_i = old_dirs[i].dot(r) * ro[i]
                    r.add_(old_stps[i], alpha=al[i] - be_i)

            if prev_flat_grad is None:
                prev_flat_grad = flat_grad.clone(memory_format=torch.contiguous_format)
            else:
                prev_flat_grad.copy_(flat_grad)
            prev_loss = loss

            ############################################################
            # compute step length
            ############################################################
            # reset initial guess for step size
            if state['n_iter'] == 1:
                t = min(1., 1. / flat_grad.abs().sum()) * lr
            else:
                t = lr

            # directional derivative
            gtd = flat_grad.dot(d)  # g * d

            # directional derivative is below tolerance
            if gtd > -tolerance_change:
                break

            # optional line search: user function
            ls_func_evals = 0
            if line_search_fn is not None:
                # perform line search, using user function
                if line_search_fn != "strong_wolfe":
                    raise RuntimeError("only 'strong_wolfe' is supported")
                else:
                    x_init = self._clone_param()

                    def obj_func(x, t, d):
                        return self._directional_evaluate(closure, x, t, d)

                    loss, flat_grad, t, ls_func_evals = _strong_wolfe(
                        obj_func, x_init, t, d, loss, flat_grad, gtd)
                self._add_grad(t, d)
                opt_cond = flat_grad.abs().max() <= tolerance_grad
            else:
                # no line search, simply move with fixed-step
                self._add_grad(t, d)
                if n_iter != max_iter:
                    # re-evaluate function only if not in last iteration
                    # the reason we do this: in a stochastic setting,
                    # no use to re-evaluate that function here
                    with torch.enable_grad():
                        loss = float(closure())
                    flat_grad = self._gather_flat_grad()
                    opt_cond = flat_grad.abs().max() <= tolerance_grad
                    ls_func_evals = 1

            # update func eval
            current_evals += ls_func_evals
            state['func_evals'] += ls_func_evals

            ############################################################
            # check conditions
            ############################################################
            if n_iter == max_iter:
                break

            if current_evals >= max_eval:
                break

            # optimal condition
            if opt_cond:
                break

            # lack of progress
            if d.mul(t).abs().max() <= tolerance_change:
                break

            if abs(loss - prev_loss) < tolerance_change:
                break

        state['d'] = d
        state['t'] = t
        state['old_dirs'] = old_dirs
        state['old_stps'] = old_stps
        state['ro'] = ro
        state['H_diag'] = H_diag
        state['prev_flat_grad'] = prev_flat_grad
        state['prev_loss'] = prev_loss

        return orig_loss

    def _clone_param(self):
        return [p.clone(memory_format=torch.contiguous_format) for p in self._params]

    def _set_param(self, params_data):
        for p, pdata in zip(self._params, params_data):
            p.copy_(pdata)

    def _directional_evaluate(self, closure, x, t, d):
        self._add_grad(t, d)
        with torch.enable_grad():
            loss = float(closure())
        flat_grad = self._gather_flat_grad()
        self._set_param(x)
        return loss, flat_grad

class Shampoo(Optimizer):
    """Shampoo optimizer for deep learning"""
    def __init__(self, params, lr=0.03, eps=1e-4, update_freq=100, weight_decay=0):
        defaults = dict(lr=lr, eps=eps, update_freq=update_freq, weight_decay=weight_decay)
        super().__init__(params, defaults)

    def step(self, closure=None):
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue

                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError('Shampoo does not support sparse gradients')

                state = self.state[p]

                # state initialization
                if len(state) == 0:
                    state['step'] = 0
                    if len(p.shape) >= 2:
                        state['G_l'] = torch.eye(p.shape[0], device=p.device, dtype=p.dtype)
                        state['G_r'] = torch.eye(p.shape[1], device=p.device, dtype=p.dtype)
                    else:
                        state['G'] = torch.zeros_like(p)

                state['step'] += 1

                if group['weight_decay'] != 0:
                    grad = grad.add(p.data, alpha=group['weight_decay'])

                if len(p.shape) >= 2:
                    # matrix case - use left and right preconditioning
                    G_l = state['G_l']
                    G_r = state['G_r']
                    
                    # update preconditioners
                    G_l.add_(torch.mm(grad, grad.t()))
                    G_r.add_(torch.mm(grad.t(), grad))
                    
                    if state['step'] % group['update_freq'] == 0:
                        # compute matrix square roots
                        try:
                            G_l_sqrt_inv = torch.linalg.inv(torch.linalg.matrix_power(G_l, 0.25))
                            G_r_sqrt_inv = torch.linalg.inv(torch.linalg.matrix_power(G_r, 0.25))
                            
                            # preconditioning
                            search_direction = torch.mm(torch.mm(G_l_sqrt_inv, grad), G_r_sqrt_inv)
                        except:
                            # fallback to regular gradient if matrix operations fail
                            search_direction = grad
                    else:
                        search_direction = grad
                else:
                    # vector case - use diagonal preconditioning
                    G = state['G']
                    G.add_(grad ** 2)
                    
                    search_direction = grad / (torch.sqrt(G) + group['eps'])

                p.data.add_(search_direction, alpha=-group['lr'])

        return loss

class AdaHessian(Optimizer):
    """AdaHessian optimizer using second-order information"""
    def __init__(self, params, lr=0.15, betas=(0.9, 0.999), eps=1e-4, weight_decay=0, 
                 hessian_power=1, update_each=1, n_samples=1, avg_conv_kernel=False):
        if not 0.0 <= lr:
            raise ValueError(f"Invalid learning rate: {lr}")
        if not 0.0 <= eps:
            raise ValueError(f"Invalid epsilon value: {eps}")
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError(f"Invalid beta parameter at index 0: {betas[0]}")
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError(f"Invalid beta parameter at index 1: {betas[1]}")
        
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay,
                       hessian_power=hessian_power, update_each=update_each, 
                       n_samples=n_samples, avg_conv_kernel=avg_conv_kernel)
        super().__init__(params, defaults)

    def get_trace(self, gradsH):
        """Compute trace of Hessian"""
        trace = 0.
        for grad, Hv in gradsH:
            trace += torch.sum(grad * Hv).cpu().item()
        return trace

    def step(self, closure=None):
        loss = None
        if closure is not None:
            loss = closure()

        params = []
        grads = []
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is not None:
                    params.append(p)
                    grads.append(p.grad)

        # compute Hessian diagonal approximation
        device = grads[0].device
        zv = [torch.randint_like(p, high=2, device=device) for p in params]
        # rademacher random variables
        for z in zv:
            z[z == 0] = -1

        h_zv = torch.autograd.grad(grads, params, grad_outputs=zv, only_inputs=True, retain_graph=False)

        for group in self.param_groups:
            for i, p in enumerate(group['params']):
                if p.grad is None:
                    continue

                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError('AdaHessian does not support sparse gradients')

                state = self.state[p]

                # state initialization
                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p.data)
                    state['exp_hessian_diag_sq'] = torch.zeros_like(p.data)

                exp_avg, exp_hessian_diag_sq = state['exp_avg'], state['exp_hessian_diag_sq']
                beta1, beta2 = group['betas']
                state['step'] += 1

                # add weight decay
                if group['weight_decay'] != 0:
                    grad = grad.add(p.data, alpha=group['weight_decay'])

                # update biased first moment estimate
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)

                # get Hessian diagonal element
                hut_trace = h_zv[i] * zv[i]
                
                # update biased second raw moment estimate
                exp_hessian_diag_sq.mul_(beta2).addcmul_(hut_trace, hut_trace, value=1 - beta2)

                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']

                # scale learning rate
                scaled_lr = group['lr'] * math.sqrt(bias_correction2) / bias_correction1

                # apply Hessian power
                hessian_powered = exp_hessian_diag_sq ** group['hessian_power']

                p.data.addcdiv_(exp_avg, hessian_powered.sqrt() + group['eps'], value=-scaled_lr)

        return loss

class KFAC(Optimizer):
    """K-FAC (Kronecker-Factored Approximate Curvature) optimizer"""
    def __init__(self, params, lr=0.001, momentum=0.9, weight_decay=0, damping=1e-3, 
                 TCov=10, TInv=100, batch_averaged=True):
        defaults = dict(lr=lr, momentum=momentum, weight_decay=weight_decay, 
                       damping=damping, TCov=TCov, TInv=TInv, batch_averaged=batch_averaged)
        super().__init__(params, defaults)
        
        self.steps = 0

    def step(self, closure=None):
        """Perform a single optimization step"""
        loss = None
        if closure is not None:
            loss = closure()

        self.steps += 1

        for group in self.param_groups:
            # update covariances and inverses
            if self.steps % group['TCov'] == 0:
                self._update_cov(group)
            
            if self.steps % group['TInv'] == 0:
                self._update_inv(group)

            # apply updates
            for p in group['params']:
                if p.grad is None:
                    continue

                state = self.state[p]
                
                if len(state) == 0:
                    state['momentum_buffer'] = torch.zeros_like(p.data)

                buf = state['momentum_buffer']
                
                # natural gradient computation (simplified)
                if len(p.shape) == 2:  # linear layer
                    grad = p.grad.data
                    if group['weight_decay'] != 0:
                        grad = grad.add(p.data, alpha=group['weight_decay'])
                    
                    # apply KFAC preconditioning (simplified)
                    natural_grad = grad  # placeholder - full KFAC needs input/output covariances
                else:
                    natural_grad = p.grad.data

                # momentum
                buf.mul_(group['momentum']).add_(natural_grad)
                
                # update parameters
                p.data.add_(buf, alpha=-group['lr'])

        return loss

    def _update_cov(self, group):
        """Update covariance matrices"""
        # simplified - full implementation needs module hooks
        pass

    def _update_inv(self, group):
        """Update inverse covariance matrices"""
        # simplified - full implementation needs eigendecomposition
        pass

class NaturalGradient(Optimizer):
    """Natural Gradient optimizer"""
    def __init__(self, params, lr=0.01, alpha=0.95, eps=1e-4):
        defaults = dict(lr=lr, alpha=alpha, eps=eps)
        super().__init__(params, defaults)

    def step(self, closure=None):
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue

                grad = p.grad.data
                state = self.state[p]

                if len(state) == 0:
                    state['step'] = 0
                    state['v'] = torch.zeros_like(p.data)

                v = state['v']
                alpha, eps = group['alpha'], group['eps']

                state['step'] += 1

                # exponential moving average of squared gradients
                v.mul_(alpha).add_(grad ** 2, alpha=1 - alpha)

                # natural gradient step
                natural_grad = grad / (torch.sqrt(v) + eps)
                p.data.add_(natural_grad, alpha=-group['lr'])

        return loss

def _strong_wolfe(obj_func, x, t, d, f, g, gtd, c1=1e-4, c2=0.9, tolerance_change=1e-9, max_ls=25):
    """Strong Wolfe line search"""
    # This is a simplified version - full implementation would be more complex
    d_norm = d.abs().max()
    g = g.clone(memory_format=torch.contiguous_format)
    
    # evaluate objective and gradient at the initial step length
    f_new, g_new = obj_func(x, t, d)
    ls_func_evals = 1
    gtd_new = g_new.dot(d)

    # check first Wolfe condition
    if f_new > (f + c1 * t * gtd) or ls_func_evals >= max_ls:
        return f_new, g_new, t, ls_func_evals

    # check second Wolfe condition
    if abs(gtd_new) <= -c2 * gtd:
        return f_new, g_new, t, ls_func_evals

    # if we get here, return the current point
    return f_new, g_new, t, ls_func_evals

__all__ = ['LBFGS', 'Shampoo', 'AdaHessian', 'KFAC', 'NaturalGradient']



"""
Margin-based loss functions.
"""

import torch
import torch.nn as nn
from ...utils.registry import register_loss

@register_loss("tripletloss")
class TripletLoss(nn.TripletMarginLoss):
    """Enhanced TripletLoss."""
    pass

@register_loss("contrastiveloss")
class ContrastiveLoss(nn.Module):
    """Contrastive Loss for metric learning."""
    
    def __init__(self, margin=1.0, reduction='mean'):
        super().__init__()
        self.margin = margin
        self.reduction = reduction
    
    def forward(self, input1, input2, target):
        distances = F.pairwise_distance(input1, input2)
        losses = (1 - target) * torch.pow(distances, 2) + \
                 target * torch.pow(torch.clamp(self.margin - distances, min=0.0), 2)
        
        if self.reduction == 'mean':
            return losses.mean()
        elif self.reduction == 'sum':
            return losses.sum()
        else:
            return losses

# Placeholder classes
class AngularLoss(nn.Module):
    def forward(self, x, y): return torch.tensor(0.0)

class ArcFaceLoss(nn.Module):
    def forward(self, x, y): return torch.tensor(0.0)

class CosFaceLoss(nn.Module):
    def forward(self, x, y): return torch.tensor(0.0)



"""
Optimizers specialized for distributed training.
"""

import torch
import torch.optim as optim
from typing import Optional, Callable, Union, List, Dict, Any
from ...utils.registry import register_optimizer

@register_optimizer("lars")
class LARS(optim.Optimizer):
    """
    LARS: Layer-wise Adaptive Rate Scaling.
    
    Reference: https://arxiv.org/abs/1708.03888
    """
    
    def __init__(
        self,
        params: Union[List[torch.Tensor], Dict[str, Any]],
        lr: float = 1.0,
        momentum: float = 0.9,
        weight_decay: float = 1e-4,
        trust_coefficient: float = 1e-3,
        eps: float = 1e-8,
        **kwargs
    ):
        defaults = dict(
            lr=lr, momentum=momentum, weight_decay=weight_decay,
            trust_coefficient=trust_coefficient, eps=eps
        )
        super().__init__(params, defaults)
    
    def step(self, closure: Optional[Callable] = None):
        """Performs a single optimization step."""
        loss = None
        if closure is not None:
            loss = closure()
        
        for group in self.param_groups:
            weight_decay = group['weight_decay']
            momentum = group['momentum']
            trust_coeff = group['trust_coefficient']
            
            for p in group['params']:
                if p.grad is None:
                    continue
                
                param_state = self.state[p]
                d_p = p.grad.data
                
                if weight_decay != 0:
                    d_p = d_p.add(p.data, alpha=weight_decay)
                
                weight_norm = torch.norm(p.data)
                grad_norm = torch.norm(d_p)
                
                # Compute local learning rate
                if weight_norm > 0 and grad_norm > 0:
                    local_lr = trust_coeff * weight_norm / (grad_norm + group['eps'])
                else:
                    local_lr = 1.0
                
                local_lr = min(local_lr, 1.0)  # Clip to maximum of 1.0
                
                if momentum != 0:
                    if 'momentum_buffer' not in param_state:
                        buf = param_state['momentum_buffer'] = torch.zeros_like(p.data)
                        buf.mul_(momentum).add_(d_p)
                    else:
                        buf = param_state['momentum_buffer']
                        buf.mul_(momentum).add_(d_p)
                    
                    d_p = buf
                
                # Apply update with local learning rate
                p.data.add_(d_p, alpha=-group['lr'] * local_lr)
        
        return loss



"""
Computer vision loss functions.
"""

from .segmentation import DiceLoss, IoULoss, TverskyLoss, FocalTverskyLoss, LovaszLoss, BoundaryLoss
from .detection import FocalDetectionLoss, GIoULoss, DIoULoss, CIoULoss, EIoULoss, AlphaIoULoss
from .super_resolution import PerceptualLoss, SSIMLoss, MSSSIMLoss, LPIPSLoss, VGGLoss
from .style_transfer import StyleLoss, ContentLoss, TotalVariationLoss

__all__ = [
    "DiceLoss", "IoULoss", "TverskyLoss", "FocalTverskyLoss", "LovaszLoss", "BoundaryLoss",
    "FocalDetectionLoss", "GIoULoss", "DIoULoss", "CIoULoss", "EIoULoss", "AlphaIoULoss",
    "PerceptualLoss", "SSIMLoss", "MSSSIMLoss", "LPIPSLoss", "VGGLoss",
    "StyleLoss", "ContentLoss", "TotalVariationLoss",
]



import torch.nn as nn
import torch

class TukeyLoss(nn.Module):
    def forward(self, x, y): return torch.tensor(0.0)

class CauchyLoss(nn.Module):
    def forward(self, x, y): return torch.tensor(0.0)

class WelschLoss(nn.Module):
    def forward(self, x, y): return torch.tensor(0.0)

class FairLoss(nn.Module):
    def forward(self, x, y): return torch.tensor(0.0)



import torch.nn as nn
import torch

# Import generative losses from domain_specific module where they are implemented
from ..domain_specific import (
    GANLoss, WassersteinLoss, HingeGANLoss, LeastSquaresGANLoss, RelativistGANLoss,
    ELBOLoss, BetaVAELoss, BetaTCVAELoss, FactorVAELoss,
    DDPMLoss, DDIMLoss, ScoreMatchingLoss
)

__all__ = [
    "GANLoss", "WassersteinLoss", "HingeGANLoss", "LeastSquaresGANLoss", "RelativistGANLoss",
    "ELBOLoss", "BetaVAELoss", "BetaTCVAELoss", "FactorVAELoss",
    "DDPMLoss", "DDIMLoss", "ScoreMatchingLoss"
]



"""
Torchium Benchmarks Package

This package contains performance benchmarks for optimizers and loss functions.
"""

from .optimizer_benchmark import OptimizerBenchmark
from .loss_benchmark import LossBenchmark
from .convergence_benchmark import ConvergenceBenchmark
from .memory_benchmark import MemoryBenchmark

__all__ = [
    'OptimizerBenchmark',
    'LossBenchmark', 
    'ConvergenceBenchmark',
    'MemoryBenchmark'
] 